# -*- coding: utf-8 -*-
# Generated by Django 1.11.24 on 2019-09-18 20:37


from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0075_auto_20190916_1030'),
    ]

    operations = [
        migrations.AddField(
            model_name='enterprisecustomer',
            name='enable_portal_reporting_config_screen',
            field=models.BooleanField(default=False, help_text='Specifies whether to allow access to the reporting configurations screen in the admin portal.'),
        ),
        migrations.AddField(
            model_name='historicalenterprisecustomer',
            name='enable_portal_reporting_config_screen',
            field=models.BooleanField(default=False, help_text='Specifies whether to allow access to the reporting configurations screen in the admin portal.'),
        ),
        migrations.AlterField(
            model_name='enterprisecustomer',
            name='learner_portal_hostname',
            field=models.CharField(blank=True, default='', help_text='Hostname of the enterprise learner portal, e.g. bestrun.edx.org.', max_length=255),
        ),
        migrations.AlterField(
            model_name='historicalenterprisecustomer',
            name='learner_portal_hostname',
            field=models.CharField(blank=True, default='', help_text='Hostname of the enterprise learner portal, e.g. bestrun.edx.org.', max_length=255),
        ),
    ]



"""
Admin integration for configuring Blackboard app to communicate with Blackboard systems.
"""

from django.contrib import admin

from integrated_channels.blackboard.models import BlackboardEnterpriseCustomerConfiguration


@admin.register(BlackboardEnterpriseCustomerConfiguration)
class BlackboardEnterpriseCustomerConfigurationAdmin(admin.ModelAdmin):
    """
    Django admin model for BlackEnterpriseCustomerConfiguration.
    """
    list_display = (
        "enterprise_customer_name",
        "client_id",
        "client_secret",
        "blackboard_base_url",
    )

    readonly_fields = (
        "enterprise_customer_name",
        "refresh_token",
    )

    search_fields = ("enterprise_customer_name",)

    class Meta:
        model = BlackboardEnterpriseCustomerConfiguration

    def enterprise_customer_name(self, obj):
        """
        Returns: the name for the attached EnterpriseCustomer.

        Args:
            obj: The instance of BlackboardEnterpriseCustomerConfiguration
                being rendered with this admin form.
        """
        return obj.enterprise_customer.name



"""
Perform health checks om the services enterprise is dependant on.
"""
import traceback

from requests.exceptions import ConnectionError, Timeout  # pylint: disable=redefined-builtin
from slumber.exceptions import HttpServerError, SlumberBaseException

from enterprise.api_client.discovery import NoAuthDiscoveryClient
from enterprise.api_client.ecommerce import NoAuthEcommerceClient
from enterprise.api_client.enterprise_catalog import NoAuthEnterpriseCatalogClient
from enterprise.api_client.lms import NoAuthLMSClient
from enterprise.heartbeat.exceptions import (
    DiscoveryNotAvailable,
    EcommerceNotAvailable,
    EnterpriseCatalogNotAvailable,
    LMSNotAvailable,
)


def check_lms():
    """
    Check if LMS service is up and running and accessible via API.

    Raises:
        (LMSNotAvailable): raised if LMS service is not accessible for some reason.

    Returns:
        (str, str): A tuple containing service name and a message.
    """
    client = NoAuthLMSClient()
    try:
        client.get_health()
    except HttpServerError as error:
        raise LMSNotAvailable('Service is down.', traceback.format_exc()) from error
    except (ConnectionError, Timeout) as error:
        raise LMSNotAvailable('Service is not accessible.', traceback.format_exc()) from error
    except SlumberBaseException as exception:
        raise LMSNotAvailable(
            'An error occurred while checking service status.', traceback.format_exc()
        ) from exception

    return 'Learning Management System (LMS)', 'Service is up and running.'


def check_ecommerce():
    """
    Check if E-Commerce service is up and running and accessible via API.

    Raises:
        (EcommerceNotAvailable): raised if LMS service is not accessible for some reason.

    Returns:
        (str, str): A tuple containing service name and a message.
    """
    client = NoAuthEcommerceClient()
    try:
        client.get_health()
    except HttpServerError as error:
        raise EcommerceNotAvailable('Service is down.', traceback.format_exc()) from error
    except (ConnectionError, Timeout) as error:
        raise EcommerceNotAvailable('Service is not accessible.', traceback.format_exc()) from error
    except SlumberBaseException as exception:
        raise EcommerceNotAvailable(
            'An error occurred while checking service status.', traceback.format_exc()
        ) from exception

    return 'E-Commerce', 'Service is up and running.'


def check_discovery():
    """
    Check if course discovery service is up and running and accessible via API.

    Raises:
        (DiscoveryNotAvailable): raised if LMS service is not accessible for some reason.

    Returns:
        (str, str): A tuple containing service name and a message.
    """
    client = NoAuthDiscoveryClient()
    try:
        client.get_health()
    except HttpServerError as error:
        raise DiscoveryNotAvailable('Service is down.', traceback.format_exc()) from error
    except (ConnectionError, Timeout) as error:
        raise DiscoveryNotAvailable('Service is not accessible.', traceback.format_exc()) from error
    except SlumberBaseException as exception:
        raise DiscoveryNotAvailable(
            'An error occurred while checking service status.', traceback.format_exc()
        ) from exception

    return 'Course Discovery', 'Service is up and running.'


def check_enterprise_catalog():
    """
    Check if enterprise catalog service is up and running and accessible via API.

    Raises:
        (EnterpriseCatalogNotAvailable): raised if LMS service is not accessible for some reason.

    Returns:
        (str, str): A tuple containing service name and a message.
    """
    client = NoAuthEnterpriseCatalogClient()
    try:
        client.get_health()
    except HttpServerError as error:
        raise EnterpriseCatalogNotAvailable('Service is down.', traceback.format_exc()) from error
    except (ConnectionError, Timeout) as error:
        raise EnterpriseCatalogNotAvailable('Service is not accessible.', traceback.format_exc()) from error
    except SlumberBaseException as exception:
        raise EnterpriseCatalogNotAvailable(
            'An error occurred while checking service status.', traceback.format_exc()
        ) from exception

    return 'Enterprise Catalog', 'Service is up and running.'



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2017-03-17 14:02


import django.utils.timezone
from django.db import migrations, models

import model_utils.fields


class Migration(migrations.Migration):

    dependencies = [
        ('sap_success_factors', '0002_auto_20170224_1545'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='catalogtransmissionaudit',
            name='end',
        ),
        migrations.RemoveField(
            model_name='catalogtransmissionaudit',
            name='start',
        ),
        migrations.AddField(
            model_name='catalogtransmissionaudit',
            name='created',
            field=model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created'),
        ),
        migrations.AddField(
            model_name='catalogtransmissionaudit',
            name='modified',
            field=model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified'),
        ),
        migrations.AddField(
            model_name='historicalsapsuccessfactorsenterprisecustomerconfiguration',
            name='user_type',
            field=models.CharField(choices=[('user', 'User'), ('admin', 'Admin')], default='user', max_length=20, verbose_name='SAP User Type'),
        ),
        migrations.AddField(
            model_name='sapsuccessfactorsenterprisecustomerconfiguration',
            name='user_type',
            field=models.CharField(choices=[('user', 'User'), ('admin', 'Admin')], default='user', max_length=20, verbose_name='SAP User Type'),
        ),
    ]



"""
Management command for assigning enterprise roles to existing enterprise users.
"""

import logging
from time import sleep

from django.apps import apps
from django.contrib import auth
from django.core.management.base import BaseCommand, CommandError

from enterprise.constants import (
    EDX_ORG_NAME,
    ENTERPRISE_ADMIN_ROLE,
    ENTERPRISE_CATALOG_ADMIN_ROLE,
    ENTERPRISE_DATA_API_ACCESS_GROUP,
    ENTERPRISE_ENROLLMENT_API_ACCESS_GROUP,
    ENTERPRISE_ENROLLMENT_API_ADMIN_ROLE,
    ENTERPRISE_LEARNER_ROLE,
    ENTERPRISE_OPERATOR_ROLE,
    OAUTH2_PROVIDER_APPLICATION_MODEL,
)
from enterprise.models import (
    EnterpriseCustomerUser,
    EnterpriseFeatureRole,
    EnterpriseFeatureUserRoleAssignment,
    SystemWideEnterpriseRole,
    SystemWideEnterpriseUserRoleAssignment,
)

LOGGER = logging.getLogger(__name__)
User = auth.get_user_model()


class Command(BaseCommand):
    """
    Management command for assigning enterprise roles to existing enterprise users.

    Example usage:
        $ ./manage.py assign_enterprise_user_roles --role=enterprise_admin
        $ ./manage.py assign_enterprise_user_roles --role=enterprise_learner
        $ ./manage.py assign_enterprise_user_roles --role=enterprise_openedx_operator
        $ ./manage.py assign_enterprise_user_roles --role=catalog_admin
        $ ./manage.py assign_enterprise_user_roles --role=enrollment_api_admin
    """
    help = 'Assigns enterprise roles to existing enterprise users.'

    def add_arguments(self, parser):
        """
        Entry point for subclassed commands to add custom arguments.
        """
        parser.add_argument(
            '--role',
            action='store',
            dest='role',
            default=None,
            help='Role to assign users role assignments.'
        )
        parser.add_argument(
            '--batch-limit',
            action='store',
            dest='batch_limit',
            default=100,
            help='Number of users in each batch of conditional offer migration.',
            type=int,
        )

        parser.add_argument(
            '--batch-offset',
            action='store',
            dest='batch_offset',
            default=0,
            help='Which index to start batching from.',
            type=int,
        )

        parser.add_argument(
            '--batch-sleep',
            action='store',
            dest='batch_sleep',
            default=10,
            help='How long to sleep between batches.',
            type=int,
        )

    def _get_enterprise_customer_user_ids(self):
        """
        Returns a queryset containing user ids.
        """
        return EnterpriseCustomerUser.objects.values('user_id')

    def _get_enterprise_admin_users_batch(self, start, end):
        """
        Returns a batched queryset of User objects.
        """
        LOGGER.info('Fetching new batch of enterprise admin users from indexes: %s to %s', start, end)
        return User.objects.filter(groups__name=ENTERPRISE_DATA_API_ACCESS_GROUP, is_staff=False)[start:end]

    def _get_enterprise_operator_users_batch(self, start, end):
        """
        Returns a batched queryset of User objects.
        """
        LOGGER.info('Fetching new batch of enterprise operator users from indexes: %s to %s', start, end)
        return User.objects.filter(groups__name=ENTERPRISE_DATA_API_ACCESS_GROUP, is_staff=True)[start:end]

    def _get_enterprise_customer_users_batch(self, start, end):
        """
        Returns a batched queryset of EnterpriseCustomerUser objects.
        """
        LOGGER.info('Fetching new batch of enterprise customer users from indexes: %s to %s', start, end)
        return User.objects.filter(pk__in=self._get_enterprise_customer_user_ids())[start:end]

    def _get_enterprise_enrollment_api_admin_users_batch(self, start, end):     # pylint: disable=invalid-name
        """
        Returns a batched queryset of User objects.
        """
        LOGGER.info('Fetching new batch of enterprise enrollment admin users from indexes: %s to %s', start, end)
        return User.objects.filter(groups__name=ENTERPRISE_ENROLLMENT_API_ACCESS_GROUP, is_staff=False)[start:end]

    def _get_enterprise_catalog_admin_users_batch(self, start, end):
        """
        Returns a batched queryset of User objects.
        """
        Application = apps.get_model(OAUTH2_PROVIDER_APPLICATION_MODEL)     # pylint: disable=invalid-name
        LOGGER.info('Fetching new batch of enterprise catalog admin users from indexes: %s to %s', start, end)
        catalog_admin_user_ids = Application.objects.filter(
            user_id__in=self._get_enterprise_customer_user_ids()
        ).exclude(name=EDX_ORG_NAME).values('user_id')
        return User.objects.filter(pk__in=catalog_admin_user_ids)[start:end]

    def _assign_enterprise_role_to_users(self, _get_batch_method, options, is_feature_role=False):
        """
        Assigns enterprise role to users.
        """
        role_name = options['role']
        batch_limit = options['batch_limit']
        batch_sleep = options['batch_sleep']
        batch_offset = options['batch_offset']

        current_batch_index = batch_offset

        users_batch = _get_batch_method(
            batch_offset,
            batch_offset + batch_limit
        )

        role_class = SystemWideEnterpriseRole
        role_assignment_class = SystemWideEnterpriseUserRoleAssignment

        if is_feature_role:
            role_class = EnterpriseFeatureRole
            role_assignment_class = EnterpriseFeatureUserRoleAssignment

        enterprise_role = role_class.objects.get(name=role_name)
        while users_batch.count() > 0:
            for index, user in enumerate(users_batch):
                LOGGER.info(
                    'Processing user with index %s and id %s',
                    current_batch_index + index, user.id
                )
                role_assignment_class.objects.get_or_create(
                    user=user,
                    role=enterprise_role
                )

            sleep(batch_sleep)
            current_batch_index += len(users_batch)
            users_batch = _get_batch_method(
                current_batch_index,
                current_batch_index + batch_limit
            )

    def handle(self, *args, **options):
        """
        Entry point for managment command execution.
        """
        LOGGER.info('Starting assigning enterprise roles to users!')

        role = options['role']
        if role == ENTERPRISE_ADMIN_ROLE:
            # Assign admin role to non-staff users with enterprise data api access.
            self._assign_enterprise_role_to_users(self._get_enterprise_admin_users_batch, options)
        elif role == ENTERPRISE_OPERATOR_ROLE:
            # Assign operator role to staff users with enterprise data api access.
            self._assign_enterprise_role_to_users(self._get_enterprise_operator_users_batch, options)
        elif role == ENTERPRISE_LEARNER_ROLE:
            # Assign enterprise learner role to enterprise customer users.
            self._assign_enterprise_role_to_users(self._get_enterprise_customer_users_batch, options)
        elif role == ENTERPRISE_ENROLLMENT_API_ADMIN_ROLE:
            # Assign enterprise enrollment api admin to non-staff users with enterprise data api access.
            self._assign_enterprise_role_to_users(self._get_enterprise_enrollment_api_admin_users_batch, options, True)
        elif role == ENTERPRISE_CATALOG_ADMIN_ROLE:
            # Assign enterprise catalog admin role to users with having credentials in catalog.
            self._assign_enterprise_role_to_users(self._get_enterprise_catalog_admin_users_batch, options, True)
        else:
            raise CommandError('Please provide a valid role name. Supported roles are {admin} and {learner}'.format(
                admin=ENTERPRISE_ADMIN_ROLE,
                learner=ENTERPRISE_LEARNER_ROLE
            ))

        LOGGER.info('Successfully finished assigning enterprise roles to users!')



# -*- coding: utf-8 -*-
# Generated by Django 1.10.2 on 2016-11-14 04:34


from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0003_auto_20161104_0937'),
    ]

    operations = [
        migrations.AddField(
            model_name='enterprisecustomer',
            name='identity_provider',
            field=models.SlugField(blank=True, default=None, null=True),
        ),
        migrations.AddField(
            model_name='historicalenterprisecustomer',
            name='identity_provider',
            field=models.SlugField(blank=True, default=None, null=True),
        ),
    ]



# -*- coding: utf-8 -*-


from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0022_auto_20170720_1543'),
    ]

    operations = [
        migrations.AddField(
            model_name='enterprisecustomer',
            name='enable_audit_data_reporting',
            field=models.BooleanField(default=False, help_text='Specifies whether to pass-back audit track enrollment data through an integrated channel.'),
        ),
        migrations.AddField(
            model_name='historicalenterprisecustomer',
            name='enable_audit_data_reporting',
            field=models.BooleanField(default=False, help_text='Specifies whether to pass-back audit track enrollment data through an integrated channel.'),
        ),
    ]



"""
Celery tasks for integrated channel management commands.
"""

import time

from celery import shared_task
from celery.utils.log import get_task_logger
from edx_django_utils.monitoring import set_code_owner_attribute

from django.contrib import auth
from django.utils import timezone

from enterprise.utils import get_enterprise_customer_for_user
from integrated_channels.integrated_channel.management.commands import (
    INTEGRATED_CHANNEL_CHOICES,
    IntegratedChannelCommandUtils,
)
from integrated_channels.utils import generate_formatted_log

LOGGER = get_task_logger(__name__)
User = auth.get_user_model()


@shared_task
@set_code_owner_attribute
def transmit_content_metadata(username, channel_code, channel_pk):
    """
    Task to send content metadata to each linked integrated channel.

    Arguments:
        username (str): The username of the User to be used for making API requests to retrieve content metadata.
        channel_code (str): Capitalized identifier for the integrated channel.
        channel_pk (str): Primary key for identifying integrated channel.

    """
    start = time.time()
    api_user = User.objects.get(username=username)
    integrated_channel = INTEGRATED_CHANNEL_CHOICES[channel_code].objects.get(pk=channel_pk)
    LOGGER.info('[Integrated Channel] Content metadata transmission started.'
                ' Configuration: {configuration}'.format(configuration=integrated_channel))
    try:
        integrated_channel.transmit_content_metadata(api_user)
    except Exception:  # pylint: disable=broad-except
        LOGGER.exception(
            '[Integrated Channel] Transmission of content metadata failed.'
            ' ChannelCode: {channel_code}, ChannelId: {channel_id}, Username: {user}'.format(
                user=username,
                channel_code=channel_code,
                channel_id=channel_pk
            ))
    duration = time.time() - start
    LOGGER.info(
        '[Integrated Channel] Content metadata transmission task finished. Configuration: {configuration},'
        'Duration: {duration}'.format(
            configuration=integrated_channel,
            duration=duration
        ))


@shared_task
@set_code_owner_attribute
def transmit_learner_data(username, channel_code, channel_pk):
    """
    Task to send learner data to each linked integrated channel.

    Arguments:
        username (str): The username of the User to be used for making API requests for learner data.
        channel_code (str): Capitalized identifier for the integrated channel
        channel_pk (str): Primary key for identifying integrated channel

    """
    start = time.time()
    api_user = User.objects.get(username=username)
    integrated_channel = INTEGRATED_CHANNEL_CHOICES[channel_code].objects.get(pk=channel_pk)
    generate_formatted_log(
        'Batch processing learners for integrated channel. Configuration: {configuration}'.format(
            configuration=integrated_channel,
        ),
        channel_name=channel_code,
        enterprise_customer_identifier=api_user.username
    )

    # Note: learner data transmission code paths don't raise any uncaught exception, so we don't need a broad
    # try-except block here.
    integrated_channel.transmit_learner_data(api_user)

    duration = time.time() - start
    generate_formatted_log(
        'Batch learner data transmission task finished. Configuration: {configuration},'
        ' Duration: {duration}'.format(
            configuration=integrated_channel,
            duration=duration
        ),
        channel_name=channel_code,
        enterprise_customer_identifier=api_user.username
    )


@shared_task
@set_code_owner_attribute
def transmit_single_learner_data(username, course_run_id):
    """
    Task to send single learner data to each linked integrated channel.

    Arguments:
        username (str): The username of the learner whose data it should send.
        course_run_id (str): The course run id of the course it should send data for.
    """
    user = User.objects.get(username=username)
    LOGGER.info('[Integrated Channel] Single learner data transmission started.'
                ' Course: {course_run}, Username: {username}'.format(
                    course_run=course_run_id,
                    username=username))
    enterprise_customer = get_enterprise_customer_for_user(user)
    channel_utils = IntegratedChannelCommandUtils()
    # Transmit the learner data to each integrated channelStarting Export
    for channel in channel_utils.get_integrated_channels(
            {'channel': None, 'enterprise_customer': enterprise_customer.uuid}
    ):
        integrated_channel = INTEGRATED_CHANNEL_CHOICES[channel.channel_code()].objects.get(pk=channel.pk)
        LOGGER.info(
            '[Integrated Channel] Processing learner for transmission. Configuration: {configuration},'
            ' User: {user_id}'.format(
                configuration=integrated_channel,
                user_id=user.id))
        integrated_channel.transmit_single_learner_data(
            learner_to_transmit=user,
            course_run_id=course_run_id,
            completed_date=timezone.now(),
            grade='Pass',
            is_passing=True
        )


@shared_task
@set_code_owner_attribute
def transmit_single_subsection_learner_data(username, course_run_id, subsection_id, grade):
    """
    Task to send a single assessment level learner data record to each linked integrated channel. This task is fired off
    when an enterprise learner completes a subsection of their course, and as such only sends the data for that sub-
    section.

    Arguments:
        username (str): The username of the learner whose data it should send.
        course_run_id  (str): The course run id of the course it should send data for.
        subsection_id (str): The subsection id that the learner completed and whose grades are being reported.
        grade (str): The grade received, used to ensure we are not sending duplicate transmissions.
    """
    start = time.time()
    user = User.objects.get(username=username)
    enterprise_customer = get_enterprise_customer_for_user(user)
    channel_utils = IntegratedChannelCommandUtils()

    # Transmit the learner data to each integrated channelStarting Export
    for channel in channel_utils.get_integrated_channels(
            {'channel': None, 'enterprise_customer': enterprise_customer.uuid}
    ):
        integrated_channel = INTEGRATED_CHANNEL_CHOICES[channel.channel_code()].objects.get(pk=channel.pk)
        integrated_channel.transmit_single_subsection_learner_data(
            learner_to_transmit=user,
            course_run_id=course_run_id,
            grade=grade,
            subsection_id=subsection_id
        )

    duration = time.time() - start
    LOGGER.info(
        '[Integrated Channel] Single learner data transmission task finished.'
        ' Course: {course_run}, Duration: {duration}, Username: {username}'.format(
            username=username,
            course_run=course_run_id,
            duration=duration))


@shared_task
@set_code_owner_attribute
def transmit_subsection_learner_data(username, channel_code, channel_pk):
    """
    Task to send assessment level learner data to each linked integrated channel.

    Arguments:
        username (str): The username of the User to be used for making API requests for learner data.
        channel_code (str): Capitalized identifier for the integrated channel
        channel_pk (str): Primary key for identifying integrated channel
    """
    start = time.time()
    api_user = User.objects.get(username=username)
    integrated_channel = INTEGRATED_CHANNEL_CHOICES[channel_code].objects.get(pk=channel_pk)
    generate_formatted_log(
        'Batch processing assessment level reporting for integrated channel. Configuration: {configuration}'.format(
            configuration=integrated_channel,
        ),
        channel_name=channel_code,
        enterprise_customer_identifier=api_user.username
    )

    # Exceptions during transmission are caught and saved within the audit so no need to try/catch here
    integrated_channel.transmit_subsection_learner_data(api_user)
    duration = time.time() - start
    LOGGER.info(
        '[Integrated Channel] Bulk learner data transmission task finished.'
        'Duration: {duration}, Username: {username}'.format(
            username=username,
            duration=duration)
    )


@shared_task
@set_code_owner_attribute
def unlink_inactive_learners(channel_code, channel_pk):
    """
    Task to unlink inactive learners of provided integrated channel.

    Arguments:
        channel_code (str): Capitalized identifier for the integrated channel
        channel_pk (str): Primary key for identifying integrated channel

    """
    start = time.time()
    integrated_channel = INTEGRATED_CHANNEL_CHOICES[channel_code].objects.get(pk=channel_pk)
    LOGGER.info('Processing learners to unlink inactive users using configuration: [%s]', integrated_channel)

    # Note: learner data transmission code paths don't raise any uncaught exception, so we don't need a broad
    # try-except block here.
    integrated_channel.unlink_inactive_learners()

    duration = time.time() - start
    LOGGER.info(
        'Unlink inactive learners task for integrated channel configuration [%s] took [%s] seconds',
        integrated_channel,
        duration
    )



# Generated by Django 2.2.12 on 2020-04-21 20:42

import jsonfield.fields

import django.utils.timezone
from django.db import migrations, models

import model_utils.fields


class Migration(migrations.Migration):

    replaces = [('integrated_channel', '0001_initial'), ('integrated_channel', '0002_delete_enterpriseintegratedchannel'), ('integrated_channel', '0003_catalogtransmissionaudit_learnerdatatransmissionaudit'), ('integrated_channel', '0004_catalogtransmissionaudit_channel'), ('integrated_channel', '0005_auto_20180306_1251'), ('integrated_channel', '0006_delete_catalogtransmissionaudit'), ('integrated_channel', '0007_auto_20190925_0730')]

    initial = True

    dependencies = [
        ('enterprise', '0094_add_use_enterprise_catalog_sample'),
    ]

    operations = [
        migrations.CreateModel(
            name='LearnerDataTransmissionAudit',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('enterprise_course_enrollment_id', models.PositiveIntegerField(db_index=True)),
                ('course_id', models.CharField(max_length=255)),
                ('course_completed', models.BooleanField(default=True)),
                ('completed_timestamp', models.BigIntegerField()),
                ('instructor_name', models.CharField(blank=True, max_length=255)),
                ('grade', models.CharField(max_length=100)),
                ('status', models.CharField(max_length=100)),
                ('error_message', models.TextField(blank=True)),
                ('created', models.DateTimeField(auto_now_add=True)),
            ],
        ),
        migrations.CreateModel(
            name='ContentMetadataItemTransmission',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('integrated_channel_code', models.CharField(max_length=30)),
                ('content_id', models.CharField(max_length=255)),
                ('channel_metadata', jsonfield.fields.JSONField(dump_kwargs={'cls': jsonfield.encoder.JSONEncoder, 'separators': (',', ':')}, load_kwargs={})),
                ('enterprise_customer', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='enterprise.EnterpriseCustomer')),
            ],
            options={
                'unique_together': {('enterprise_customer', 'integrated_channel_code', 'content_id')},
            },
        ),
    ]



"""
Content metadata exporter for Moodle
"""

from logging import getLogger

from dateutil.parser import parse

from integrated_channels.integrated_channel.exporters.content_metadata import ContentMetadataExporter

LOGGER = getLogger(__name__)


class MoodleContentMetadataExporter(ContentMetadataExporter):
    """
        Moodle implementation of ContentMetadataExporter.
    """
    DATA_TRANSFORM_MAPPING = {
        'fullname': 'title',
        'shortname': 'shortname',
        'idnumber': 'key',
        'summary': 'description',
        'startdate': 'start',
        'enddate': 'end',
        'categoryid': 'categoryid',
    }

    LONG_STRING_LIMIT = 1700  # Actual maximum value we can support for any individual course
    SKIP_KEY_IF_NONE = True

    def transform_shortname(self, content_metadata_item):
        """
        We're prefixing Title to the key to make shortname a little nicer in Moodle's UI.
        But because we use key elsewhere, I mapped this to itself
        so it doesn't override all "key" references
        """
        return '{} ({})'.format(
            content_metadata_item.get('title'),
            content_metadata_item.get('key')
        )

    def transform_title(self, content_metadata_item):
        """
        Returns the course title with all organizations (partners) appended in parantheses
        """
        formatted_orgs = []
        for org in content_metadata_item.get('organizations', []):
            split_org = org.partition(': ')  # results in: ['first_part', 'delim', 'latter part']
            if split_org[2] == '':  # Occurs when the delimiter isn't present in the string.
                formatted_orgs.append(split_org[0])  # Returns the original string
            else:
                formatted_orgs.append(split_org[2])
        if not formatted_orgs:
            final_orgs = ''
        else:
            final_orgs = ' ({})'.format(', '.join(formatted_orgs))

        return '{}{}'.format(
            content_metadata_item.get('title'),
            final_orgs
        )

    def transform_description(self, content_metadata_item):
        """
        Return the course description and enrollment url as Moodle' syllabus body attribute.
        This will display in the Syllabus tab in Moodle.
        """
        enrollment_url = content_metadata_item.get('enrollment_url', None)
        base_description = '<a href={enrollment_url} target="_blank">Go to edX course page</a><br />'.format(
            enrollment_url=enrollment_url)
        full_description = content_metadata_item.get('full_description') or None
        short_description = content_metadata_item.get('short_description') or None
        if full_description and len(full_description + enrollment_url) <= self.LONG_STRING_LIMIT:
            description = "{base_description}{full_description}".format(
                base_description=base_description,
                full_description=full_description
            )
        elif short_description and len(short_description + enrollment_url) <= self.LONG_STRING_LIMIT:
            short_description = content_metadata_item.get('short_description')
            description = "{base_description}{short_description}".format(
                base_description=base_description, short_description=short_description
            )
        else:
            description = "{base_description}{title}".format(
                base_description=base_description,
                title=content_metadata_item.get('title', '')
            )

        return description

    def transform_categoryid(self, content_metadata_item):  # pylint: disable=unused-argument
        """
        Returns the Moodle category id configured in the model.
        ID 1 is Miscellaneous and is the default/basic category.
        """
        return self.enterprise_configuration.category_id or 1

    def transform_start(self, content_metadata_item):
        """
        Converts start from ISO date string to int (required for Moodle's "startdate" field)
        """
        start_date = content_metadata_item.get('start', None)
        if start_date:
            return int(parse(start_date).timestamp())
        return None

    def transform_end(self, content_metadata_item):
        """
        Converts end from ISO date string to int (required for Moodle's "enddate" field)
        """
        end_date = content_metadata_item.get('end', None)
        if end_date:
            return int(parse(end_date).timestamp())
        return None



"""
    Viewset for integrated_channels/v1/blackboard/
"""
from edx_rbac.mixins import PermissionRequiredMixin
from edx_rest_framework_extensions.auth.jwt.authentication import JwtAuthentication
from rest_framework import permissions, viewsets
from rest_framework.authentication import SessionAuthentication

from django.core.exceptions import PermissionDenied
from django.http import Http404

from enterprise.models import EnterpriseCustomerUser
from integrated_channels.blackboard.models import BlackboardEnterpriseCustomerConfiguration

from .serializers import BlackboardConfigSerializer


class BlackboardConfigurationViewSet(PermissionRequiredMixin, viewsets.ModelViewSet):
    permission_classes = (permissions.IsAuthenticated,)
    serializer_class = BlackboardConfigSerializer
    permission_required = 'enterprise.can_access_admin_dashboard'

    def get_queryset(self):
        enterprise_customer_users = EnterpriseCustomerUser.objects.filter(user_id=self.request.user.id)
        if enterprise_customer_users:
            enterprise_customers_uuids = [str(enterprise_customer_user.enterprise_customer.uuid) for
                                          enterprise_customer_user in enterprise_customer_users]
            enterprise_customer_uuid = self.request.query_params.get('enterprise_customer')

            if enterprise_customer_uuid:
                if enterprise_customer_uuid in enterprise_customers_uuids:
                    return BlackboardEnterpriseCustomerConfiguration.objects.filter(
                        enterprise_customer__uuid=enterprise_customer_uuid
                    )
                raise PermissionDenied()
            return BlackboardEnterpriseCustomerConfiguration.objects.filter(
                enterprise_customer__uuid__in=enterprise_customers_uuids
            )
        raise Http404



"""
Run health checks on all the services enterprise service is dependant on.
"""
import logging

from enterprise.heartbeat import checks
from enterprise.heartbeat.exceptions import ServiceNotAvailable

CHECKS = (
    checks.check_lms,
    checks.check_enterprise_catalog,
    checks.check_discovery,
    checks.check_ecommerce
)


LOGGER = logging.getLogger(__name__)


class Status:
    """
    Status constants for service health check.
    """
    OK = u'OK'  # pylint: disable=invalid-name
    UNAVAILABLE = u'UNAVAILABLE'  # pylint: disable=invalid-name


def run_checks():
    """
    Run health checks on all the services and return the status of each service.

    Returns:
        (bool, dict): first boolean un the tuple tells of all services are up and running or not,
            Dictionary containing the following key value pairs status: 'OK' if all the services are up
    """
    response = {
        'status': Status.OK,
        'message': 'Service is up and running.',
        'services': [],
    }

    for check_func in CHECKS:
        try:
            service_name, message = check_func()
        except ServiceNotAvailable as error:
            LOGGER.exception(
                '{error.service_name} service health check failed with message "{error.message}"'.format(error=error)
            )

            response['status'] = Status.UNAVAILABLE
            response['message'] = 'Some or all of the dependant services are down.'
            response['services'].append({
                'service': error.service_name,
                'message': error.message,
                'status': Status.UNAVAILABLE,
            })
        else:
            response['services'].append({
                'service': service_name,
                'message': message,
                'status': Status.OK,
            })

    return response['status'] == Status.OK, response



# -*- coding: utf-8 -*-
"""
Assist integrated channels with retrieving learner completion data.

Module contains resources for integrated pipelines to retrieve all the
grade and completion data for enrollments belonging to a particular
enterprise customer.
"""

from logging import getLogger

from slumber.exceptions import HttpNotFoundError

from django.apps import apps
from django.utils import timezone
from django.utils.dateparse import parse_datetime

from consent.models import DataSharingConsent
from enterprise.api_client.lms import CertificatesApiClient, CourseApiClient, GradesApiClient
from enterprise.models import EnterpriseCourseEnrollment
from integrated_channels.integrated_channel.exporters import Exporter
from integrated_channels.utils import generate_formatted_log, is_already_transmitted, parse_datetime_to_epoch_millis

LOGGER = getLogger(__name__)


class LearnerExporter(Exporter):
    """
    Base class for exporting learner completion data to integrated channels.
    """

    GRADE_AUDIT = 'Audit'
    GRADE_PASSING = 'Pass'
    GRADE_FAILING = 'Fail'
    GRADE_INCOMPLETE = 'In Progress'

    def __init__(self, user, enterprise_configuration):
        """
        Store the data needed to export the learner data to the integrated channel.

        Arguments:

        * ``user``: User instance with access to the Grades API for the Enterprise Customer's courses.
        * ``enterprise_configuration``: EnterpriseCustomerPluginConfiguration instance for the current channel.
        """
        # The Grades API and Certificates API clients require an OAuth2 access token,
        # so cache the client to allow the token to be reused. Cache other clients for
        # general reuse.
        self.grades_api = None
        self.certificates_api = None
        self.course_api = None
        self.course_enrollment_api = None

        # Cached course details data from the Course API.
        self.course_details = dict()
        super().__init__(user, enterprise_configuration)

    @property
    def grade_passing(self):
        """
        Returns the string used for a passing grade.
        """
        return self.GRADE_PASSING

    @property
    def grade_failing(self):
        """
        Returns the string used for a failing grade.
        """
        return self.GRADE_FAILING

    @property
    def grade_incomplete(self):
        """
        Returns the string used for an incomplete course grade.
        """
        return self.GRADE_INCOMPLETE

    @property
    def grade_audit(self):
        """
        Returns the string used for an audit course grade.
        """
        return self.GRADE_AUDIT

    def fetch_course_details(self, course_details, enrollment_course_id):
        """
        Using a cached value of course details to prevent duplicate work, retrieve an enrollments course details
        from the Course api. Returns None if no details could be retrieved.
        """
        # Check if first run or if the cached course ID doesn't match the enrollment course ID
        if course_details is None or course_details['course_id'] != enrollment_course_id:
            if self.course_api is None:
                self.course_api = CourseApiClient()
            course_details = self.course_api.get_course_details(enrollment_course_id)

        if course_details is None:
            return None

        return course_details

    def bulk_assessment_level_export(self):
        """
        Collect assessment level learner data for the ``EnterpriseCustomer`` where data sharing consent is granted.

        Yields a learner assessment data object for each subsection within a course under an enrollment, containing:

        * ``enterprise_enrollment``: ``EnterpriseCourseEnrollment`` object.
        * ``course_id``: The string ID of the course under the enterprise enrollment.
        * ``subsection_id``: The string ID of the subsection within the course.
        * ``grade``: string grade recorded for the learner in the course.
        """
        enrollment_queryset = EnterpriseCourseEnrollment.objects.select_related(
            'enterprise_customer_user'
        ).filter(
            enterprise_customer_user__enterprise_customer=self.enterprise_customer,
            enterprise_customer_user__active=True,
        ).order_by('course_id')

        # Create a record of each subsection from every enterprise enrollment
        for enterprise_enrollment in enrollment_queryset:
            if not LearnerExporter.has_data_sharing_consent(enterprise_enrollment):
                continue

            assessment_grade_data = self._collect_assessment_grades_data(enterprise_enrollment)

            records = self.get_learner_assessment_data_records(
                enterprise_enrollment=enterprise_enrollment,
                assessment_grade_data=assessment_grade_data,
            )
            if records:
                # There are some cases where we won't receive a record from the above
                # method; right now, that should only happen if we have an Enterprise-linked
                # user for the integrated channel, and transmission of that user's
                # data requires an upstream user identifier that we don't have (due to a
                # failure of SSO or similar). In such a case, `get_learner_data_record`
                # would return None, and we'd simply skip yielding it here.
                for record in records:
                    yield record

    def single_assessment_level_export(self, **kwargs):
        """
        Collect a single assessment level learner data for the ``EnterpriseCustomer`` where data sharing consent is
        granted.

        Yields a learner assessment data object for each subsection of the course that the learner is enrolled in,
        containing:

        * ``enterprise_enrollment``: ``EnterpriseCourseEnrollment`` object.
        * ``course_id``: The string ID of the course under the enterprise enrollment.
        * ``subsection_id``: The string ID of the subsection within the course.
        * ``grade``: string grade recorded for the learner in the course.
        """
        learner_to_transmit = kwargs.get('learner_to_transmit', None)
        TransmissionAudit = kwargs.get('TransmissionAudit', None)  # pylint: disable=invalid-name
        course_run_id = kwargs.get('course_run_id', None)
        grade = kwargs.get('grade', None)
        subsection_id = kwargs.get('subsection_id')
        enrollment_queryset = EnterpriseCourseEnrollment.objects.select_related(
            'enterprise_customer_user'
        ).filter(
            enterprise_customer_user__active=True,
            enterprise_customer_user__user_id=learner_to_transmit.id,
            course_id=course_run_id,
        ).order_by('course_id')

        # We are transmitting for a single enrollment, so grab just the one.
        enterprise_enrollment = enrollment_queryset.first()

        already_transmitted = is_already_transmitted(
            TransmissionAudit,
            enterprise_enrollment.id,
            grade,
            subsection_id
        )

        if not (TransmissionAudit and already_transmitted) and LearnerExporter.has_data_sharing_consent(
                enterprise_enrollment):

            # No caching because we're only fetching one course detail
            course_details = self.fetch_course_details(None, course_run_id)

            if course_details:
                assessment_grade_data = self._collect_assessment_grades_data(enterprise_enrollment)

                records = self.get_learner_assessment_data_records(
                    enterprise_enrollment=enterprise_enrollment,
                    assessment_grade_data=assessment_grade_data,
                )
                if records:
                    # There are some cases where we won't receive a record from the above
                    # method; right now, that should only happen if we have an Enterprise-linked
                    # user for the integrated channel, and transmission of that user's
                    # data requires an upstream user identifier that we don't have (due to a
                    # failure of SSO or similar). In such a case, `get_learner_data_record`
                    # would return None, and we'd simply skip yielding it here.
                    for record in records:
                        yield record

    @staticmethod
    def has_data_sharing_consent(enterprise_enrollment):
        """
        Helper method to determine whether an enrollment has data sharing consent or not.
        """
        consent = DataSharingConsent.objects.proxied_get(
            username=enterprise_enrollment.enterprise_customer_user.username,
            course_id=enterprise_enrollment.course_id,
            enterprise_customer=enterprise_enrollment.enterprise_customer_user.enterprise_customer
        )
        if consent.granted and not enterprise_enrollment.audit_reporting_disabled:
            return True

        return False

    def export(self, **kwargs):  # pylint: disable=R0915
        """
        Collect learner data for the ``EnterpriseCustomer`` where data sharing consent is granted.

        Yields a learner data object for each enrollment, containing:

        * ``enterprise_enrollment``: ``EnterpriseCourseEnrollment`` object.
        * ``completed_date``: datetime instance containing the course/enrollment completion date; None if not complete.
          "Course completion" occurs for instructor-paced courses when course certificates are issued, and
          for self-paced courses, when the course end date is passed, or when the learner achieves a passing grade.
        * ``grade``: string grade recorded for the learner in the course.
        """
        channel_name = kwargs.get('app_label')
        exporting_single_learner = False
        learner_to_transmit = kwargs.get('learner_to_transmit', None)
        course_run_id = kwargs.get('course_run_id', None)
        completed_date = kwargs.get('completed_date', None)
        is_passing = kwargs.get('is_passing', False)
        grade = kwargs.get('grade', None)
        skip_transmitted = kwargs.get('skip_transmitted', True)
        TransmissionAudit = kwargs.get('TransmissionAudit', None)  # pylint: disable=invalid-name
        # Fetch the consenting enrollment data, including the enterprise_customer_user.
        # Order by the course_id, to avoid fetching course API data more than we have to.
        generate_formatted_log(
            'Starting Export. CompletedDate: {completed_date}, Course: {course_run}, '
            'Grade: {grade}, IsPassing: {is_passing}, User: {user_id}'.format(
                completed_date=completed_date,
                course_run=course_run_id,
                grade=grade,
                is_passing=is_passing,
                user_id=learner_to_transmit.id if learner_to_transmit else None
            ),
            channel_name=channel_name,
            enterprise_customer_identifier=self.enterprise_customer.name
        )
        enrollment_queryset = EnterpriseCourseEnrollment.objects.select_related(
            'enterprise_customer_user'
        ).filter(
            enterprise_customer_user__enterprise_customer=self.enterprise_customer,
            enterprise_customer_user__active=True,
        )
        if learner_to_transmit and course_run_id:
            enrollment_queryset = enrollment_queryset.filter(
                course_id=course_run_id,
                enterprise_customer_user__user_id=learner_to_transmit.id,
            )
            exporting_single_learner = True
            generate_formatted_log(
                'Exporting single learner. Course: {course_run}, User: {user_id}'.format(
                    course_run=course_run_id,
                    user_id=learner_to_transmit.id
                ),
                channel_name=channel_name,
                enterprise_customer_identifier=self.enterprise_customer.name
            )
        enrollment_queryset = enrollment_queryset.order_by('course_id')

        # Fetch course details from the Course API, and cache between calls.
        course_details = None

        generate_formatted_log(
            'Beginning export of enrollments: {enrollments}.'.format(
                enrollments=list(enrollment_queryset.values()),
            ),
            channel_name=channel_name,
            enterprise_customer_identifier=self.enterprise_customer.name
        )

        for enterprise_enrollment in enrollment_queryset:
            is_audit_enrollment = enterprise_enrollment.is_audit_enrollment
            if TransmissionAudit and skip_transmitted and \
                    is_already_transmitted(TransmissionAudit, enterprise_enrollment.id, grade):
                # We've already sent a completion status for this enrollment
                generate_formatted_log(
                    'Skipping export of previously sent enterprise enrollment. '
                    'EnterpriseEnrollment: {enterprise_enrollment_id}'.format(
                        enterprise_enrollment_id=enterprise_enrollment.id
                    ),
                    channel_name=channel_name,
                    enterprise_customer_identifier=self.enterprise_customer.name
                )
                continue

            course_id = enterprise_enrollment.course_id

            # Fetch course details from Courses API
            # pylint: disable=unsubscriptable-object
            if course_details:
                generate_formatted_log(
                    'Currently exporting for course: {curr_course}, '
                    'but course details already found: {course_details}'.format(
                        curr_course=course_id,
                        course_details=course_details
                    ),
                    channel_name=channel_name,
                    enterprise_customer_identifier=self.enterprise_customer.name
                )

            if course_details is None or course_details['course_id'] != course_id:
                if self.course_api is None:
                    self.course_api = CourseApiClient()
                course_details = self.course_api.get_course_details(course_id)
                generate_formatted_log(
                    'Successfully retrieved course details for course: {}'.format(
                        course_id
                    ),
                    channel_name=channel_name,
                    enterprise_customer_identifier=self.enterprise_customer.name
                )

            if course_details is None:
                # Course not found, so we have nothing to report.
                generate_formatted_log(
                    'Course run details not found. EnterpriseEnrollment: {enterprise_enrollment_pk}, '
                    'Course: {course_id}'.format(
                        enterprise_enrollment_pk=enterprise_enrollment.pk,
                        course_id=course_id
                    ),
                    channel_name=channel_name,
                    enterprise_customer_identifier=self.enterprise_customer.name,
                    is_error=True,
                )
                continue

            if (not LearnerExporter.has_data_sharing_consent(enterprise_enrollment) or
                    enterprise_enrollment.audit_reporting_disabled):
                continue

            # For instructor-paced and not audit courses, let the certificate determine course completion
            if course_details.get('pacing') == 'instructor' and not is_audit_enrollment:
                completed_date_from_api, grade_from_api, is_passing_from_api, grade_percent = \
                    self._collect_certificate_data(enterprise_enrollment)
                generate_formatted_log(
                    'Received data from certificate api. CompletedDate: {completed_date}, Course: {course_id}, '
                    'Enterprise: {enterprise}, Grade: {grade}, IsPassing: {is_passing}, User: {user_id}'.format(
                        completed_date=completed_date_from_api,
                        grade=grade_from_api,
                        is_passing=is_passing_from_api,
                        course_id=course_id,
                        user_id=enterprise_enrollment.enterprise_customer_user.user_id,
                        enterprise=enterprise_enrollment.enterprise_customer_user.enterprise_customer.slug
                    ),
                    channel_name=channel_name,
                    enterprise_customer_identifier=self.enterprise_customer.name
                )
            # For self-paced courses, check the Grades API
            else:
                completed_date_from_api, grade_from_api, is_passing_from_api, grade_percent = \
                    self._collect_grades_data(enterprise_enrollment, course_details, is_audit_enrollment)
                generate_formatted_log(
                    'Received data from grades api. CompletedDate: {completed_date}, Course: {course_id}, '
                    'Enterprise: {enterprise}, Grade: {grade}, IsPassing: {is_passing}, User: {user_id}'.format(
                        completed_date=completed_date_from_api,
                        grade=grade_from_api,
                        is_passing=is_passing_from_api,
                        course_id=course_id,
                        user_id=enterprise_enrollment.enterprise_customer_user.user_id,
                        enterprise=enterprise_enrollment.enterprise_customer_user.enterprise_customer.slug
                    ),
                    channel_name=channel_name,
                    enterprise_customer_identifier=self.enterprise_customer.name
                )
            if exporting_single_learner and (grade != grade_from_api or is_passing != is_passing_from_api):
                enterprise_user = enterprise_enrollment.enterprise_customer_user
                generate_formatted_log(
                    'Attempt to transmit conflicting data. '
                    ' Course: {course_id}, Enterprise: {enterprise},'
                    ' EnrollmentId: {enrollment_id},'
                    ' Grade: {grade}, GradeAPI: {grade_api}, IsPassing: {is_passing},'
                    ' IsPassingAPI: {is_passing_api}, User: {user_id}'.format(
                        grade=grade,
                        is_passing=is_passing,
                        grade_api=grade_from_api,
                        is_passing_api=is_passing_from_api,
                        course_id=course_id,
                        enrollment_id=enterprise_enrollment.id,
                        user_id=enterprise_user.user_id,
                        enterprise=enterprise_user.enterprise_customer.slug
                    ),
                    channel_name=channel_name,
                    enterprise_customer_identifier=self.enterprise_customer.name,
                    is_error=True
                )
            # Apply the Single Source of Truth for Grades
            grade = grade_from_api
            completed_date = completed_date_from_api
            is_passing = is_passing_from_api
            records = self.get_learner_data_records(
                enterprise_enrollment=enterprise_enrollment,
                completed_date=completed_date,
                grade=grade,
                is_passing=is_passing,
                grade_percent=grade_percent
            )

            if records:
                # There are some cases where we won't receive a record from the above
                # method; right now, that should only happen if we have an Enterprise-linked
                # user for the integrated channel, and transmission of that user's
                # data requires an upstream user identifier that we don't have (due to a
                # failure of SSO or similar). In such a case, `get_learner_data_record`
                # would return None, and we'd simply skip yielding it here.
                for record in records:
                    yield record

    def get_learner_assessment_data_records(
            self,
            enterprise_enrollment,
            assessment_grade_data
    ):
        """
        Generate a learner assessment data transmission audit with fields properly filled in.
        """
        # pylint: disable=invalid-name
        LearnerDataTransmissionAudit = apps.get_model('integrated_channel', 'LearnerDataTransmissionAudit')
        user_subsection_audits = []
        # Create an audit for each of the subsections in the course data.
        for subsection_data in assessment_grade_data.values():
            subsection_percent_grade = subsection_data.get('grade')
            subsection_id = subsection_data.get('subsection_id')
            # Sanity check for a grade to report
            if not subsection_percent_grade or not subsection_id:
                continue

            user_subsection_audits.append(LearnerDataTransmissionAudit(
                enterprise_course_enrollment_id=enterprise_enrollment.id,
                course_id=enterprise_enrollment.course_id,
                subsection_id=subsection_id,
                grade=subsection_percent_grade,
            ))

        return user_subsection_audits

    def get_learner_data_records(
            self,
            enterprise_enrollment,
            completed_date=None,
            grade=None,
            is_passing=False,
            grade_percent=None  # pylint: disable=unused-argument
    ):
        """
        Generate a learner data transmission audit with fields properly filled in.
        """
        # pylint: disable=invalid-name
        LearnerDataTransmissionAudit = apps.get_model('integrated_channel', 'LearnerDataTransmissionAudit')
        completed_timestamp = None
        course_completed = False
        if completed_date is not None:
            completed_timestamp = parse_datetime_to_epoch_millis(completed_date)
            course_completed = is_passing

        return [
            LearnerDataTransmissionAudit(
                enterprise_course_enrollment_id=enterprise_enrollment.id,
                course_id=enterprise_enrollment.course_id,
                course_completed=course_completed,
                completed_timestamp=completed_timestamp,
                grade=grade,
            )
        ]

    def _collect_certificate_data(self, enterprise_enrollment):
        """
        Collect the learner completion data from the course certificate.

        Used for Instructor-paced courses.

        If no certificate is found, then returns the completed_date = None, grade = In Progress, on the idea that a
        certificate will eventually be generated.

        Args:
            enterprise_enrollment (EnterpriseCourseEnrollment): the enterprise enrollment record for which we need to
            collect completion/grade data

        Returns:
            completed_date: Date the course was completed, this is None if course has not been completed.
            grade: Current grade in the course.
            percent_grade: The current percent grade in the course.
            is_passing: Boolean indicating if the grade is a passing grade or not.
        """

        if self.certificates_api is None:
            self.certificates_api = CertificatesApiClient(self.user)

        course_id = enterprise_enrollment.course_id
        username = enterprise_enrollment.enterprise_customer_user.user.username

        try:
            certificate = self.certificates_api.get_course_certificate(course_id, username)
            completed_date = certificate.get('created_date')
            if completed_date:
                completed_date = parse_datetime(completed_date)
            else:
                completed_date = timezone.now()

            # For consistency with _collect_grades_data, we only care about Pass/Fail grades. This could change.
            is_passing = certificate.get('is_passing')
            percent_grade = certificate.get('grade')
            grade = self.grade_passing if is_passing else self.grade_failing

        except HttpNotFoundError:
            LOGGER.error('[Integrated Channel] Certificate data not found.'
                         ' Course: {course_id}, EnterpriseEnrollment: {enterprise_enrollment},'
                         ' Username: {username}'.format(
                             course_id=course_id,
                             username=username,
                             enterprise_enrollment=enterprise_enrollment.pk))
            completed_date = None
            grade = self.grade_incomplete
            is_passing = False
            percent_grade = None

        return completed_date, grade, is_passing, percent_grade

    def _collect_assessment_grades_data(self, enterprise_enrollment):
        """
        Collect a learner's assessment level grade data using an enterprise enrollment, from the Grades API.

        Args:
            enterprise_enrollment (EnterpriseCourseEnrollment): the enterprise enrollment record for which we need to
            collect subsection grades data
        Returns:
            Dict:
                {
                    [subsection name]: {
                        'grade_category': category,
                        'grade': percent grade,
                        'assessment_label': label,
                        'grade_point_score': points earned on the assignment,
                        'grade_points_possible': max possible points on the assignment,
                        'subsection_id': subsection module ID
                    }

                    ...
                }
        """
        if self.grades_api is None:
            self.grades_api = GradesApiClient(self.user)

        course_id = enterprise_enrollment.course_id
        username = enterprise_enrollment.enterprise_customer_user.user.username
        try:
            assessment_grades_data = self.grades_api.get_course_assessment_grades(course_id, username)
        except HttpNotFoundError:
            return {}

        assessment_grades = {}
        for grade in assessment_grades_data:
            if not grade.get('attempted'):
                continue
            assessment_grades[grade.get('subsection_name')] = {
                'grade_category': grade.get('category'),
                'grade': grade.get('percent'),
                'assessment_label': grade.get('label'),
                'grade_point_score': grade.get('score_earned'),
                'grade_points_possible': grade.get('score_possible'),
                'subsection_id': grade.get('module_id')
            }

        return assessment_grades

    def _collect_grades_data(self, enterprise_enrollment, course_details, is_audit_enrollment):
        """
        Collect the learner completion data from the Grades API.

        Used for self-paced courses.

        Args:
            enterprise_enrollment (EnterpriseCourseEnrollment): the enterprise enrollment record for which we need to
            collect completion/grade data
            course_details (dict): the course details for the course in the enterprise enrollment record.

        Returns:
            completed_date: Date the course was completed, this is None if course has not been completed.
            grade: Current grade in the course.
            is_passing: Boolean indicating if the grade is a passing grade or not.
        """
        if self.grades_api is None:
            self.grades_api = GradesApiClient(self.user)

        course_id = enterprise_enrollment.course_id
        username = enterprise_enrollment.enterprise_customer_user.user.username

        try:
            grades_data = self.grades_api.get_course_grade(course_id, username)

        except HttpNotFoundError as error:
            # Grade not found, so we have nothing to report.
            if hasattr(error, 'response'):
                response_content = error.response.json()  # pylint: disable=no-member
                if response_content.get('error_code', '') == 'user_not_enrolled':
                    # This means the user has an enterprise enrollment record but is not enrolled in the course yet
                    LOGGER.info(
                        '[Integrated Channel] User is not enrolled in the course.'
                        ' Course: {course_id}, EnterpriseEnrollment: {enterprise_enrollment},'
                        ' Username: {username}'.format(
                            course_id=course_id,
                            username=username,
                            enterprise_enrollment=enterprise_enrollment.pk))
                    return None, None, None, None

            LOGGER.error('[Integrated Channel] Grades data not found.'
                         ' Course: {course_id}, EnterpriseEnrollment: {enterprise_enrollment},'
                         ' Username: {username}'.format(
                             course_id=course_id,
                             username=username,
                             enterprise_enrollment=enterprise_enrollment.pk))
            return None, None, None, None

        # Prepare to process the course end date and pass/fail grade
        course_end_date = course_details.get('end')
        if course_end_date is not None:
            course_end_date = parse_datetime(course_end_date)
        now = timezone.now()
        is_passing = grades_data.get('passed')

        # We can consider a course complete if:
        # * the course's end date has passed
        if course_end_date is not None and course_end_date < now:
            completed_date = course_end_date
            grade = self.grade_passing if is_passing else self.grade_failing
            grade = self.grade_audit if is_audit_enrollment else grade

        # * Or, the learner has a passing grade (as of now)
        elif is_passing:
            completed_date = now
            grade = self.grade_passing

        # Otherwise, the course is still in progress
        else:
            completed_date = None
            grade = self.grade_incomplete

        percent_grade = grades_data.get('percent', None)

        return completed_date, grade, is_passing, percent_grade
